/* Author: J. Zobel, May 2001 */

#include "distrand.h"

/*
Take a continuous cumulative distribution provided in function cd, and
construct a mapping array that is designed to be accessed by a
uniformly-distributed random number.  Essentially the algorithm is:
    - Proceed along the x-axis, calculating cumulative probabilies
      at regular intervals
    - These probabilities are magnified by factor "numpoints", the quantum
      of the y-axis (which is represented in integers from 0 to numpoints)
    - For each discrete y-value, the corresponding x-value is stored.
      That is, if the y-value calculated at some x is new, record this.
    - populateregion() is called recursively to fill in regions of high
      probability.

Each distribution requires its own object of type DISTRIB.
*/
void
initdistrand(DISTRIB *dist, double cd(double))
{
    void	populateregion(DISTRIB *, int, double, int, double ());
    char	*malloc();
/* omit debug
    int		i;
*/

    dist->numpoints = NUMPOINTS;
    dist->points = (double *) malloc( sizeof(double) * (dist->numpoints+1) );

    dist->points[0] = 0.0;
    populateregion(dist, 0, 1.0, 0, cd);
    dist->points[dist->numpoints] = 1.0;

/* omit debug
    for( i=0 ; i<dist->numpoints ; i++ )
    {
	if (i%1000==0 )
	    printf("%d,%f ... ", i, dist->points[i]);
	if (i%4000==0 )
	    putchar('\n');
    }
*/
}


void
populateregion(DISTRIB *dist, int pos, double ub, int level, double cd(double))
{
    int		i, j, newpos;
    double	curr, cmlt, lb = dist->points[pos];
    double	diff = ( ub-lb ) / (double)NUMINCREMENT;

    for( i=1 ; i<=NUMINCREMENT ; i++ )
    {
	curr = lb + (double)i * diff;
	cmlt = cd(curr);
	newpos = (int)( ( cmlt ) * dist->numpoints );

	if( newpos == pos )
	{
	    /* fine, move on */
	}
	else if( newpos > pos )
	{
	    if( level<4 && newpos>pos+1 )
		/* fill in the intervening region */
	    {
		populateregion(dist, pos, curr, level+1, cd);
	    }
	    else /* interpolate */
	    {
		for( j=1 ; j<newpos-pos ; j++ )
		    dist->points[j+pos] = dist->points[pos]
				+ (double)j * ( curr-dist->points[pos] )
				  / (double)( newpos-pos );
	    }
	    pos = newpos;
	    dist->points[pos] = curr;
	}
	else /* if( newpos<pos ) */
	{
	    fprintf(stderr, "Cumulative distribution is decreasing?\n");
	    exit(1);
	}
    }
}


/* Given an integer range, returns a value in that range according
   to the provided distribution.  Assumes that random() returns
   a 31-bit number. */
int
distrand(DISTRIB *dist, int range)
{
    double	low, up;
    unsigned int rd = random();
    int		idx, rem;

    idx = rd & NUMMASK;
    low = dist->points[idx], up = dist->points[idx+1];
    rem = ( rd>>NUMBITS ) & REMMASK;

    return( (int) ( (double)range * ( low + (double)rem*(up-low)/(double)MAXREM ) ) );
}
